/*
   Copyright (C) 1998,1999 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpsim; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

//
// cod.cc
//
//  The file contains the code for reading Byte Craft's .cod
// formatted symbol files. (Technically this is a c++ file. 
// However, I use the same include file (cod.h) that's used
// by gpasm (which is written in C). Consequently
//
#include <stdio.h>
#include <iostream.h>
#include <string>
#include <string.h>

#include "pic-processor.h"
#include "symbol.h"
#include "cod.h"

static FILE *codefile = NULL;
static FILE *lstfp = NULL;
static char *directory_block_data = NULL;
static char *temp_block = NULL;
static char lstfilename[256];
#define CONFIG_ADDRESS 0x2007
static int config_word = 0xffff;

enum cod_errors
{
  COD_SUCCESS,
  COD_FILE_NOT_FOUND,
  COD_UNRECOGNIZED_PROCESSOR,
  COD_FILE_NAME_TOO_LONG,
  COD_LST_NOT_FOUND,
  COD_BAD_FILE
};

char * substr(char *a, char *b, int n)
{
  *a = 0;
  return strncat(a,b,n);

}

unsigned short get_short_int( char * buff)
{
  return ( (unsigned char)buff[0] + ((unsigned char)buff[1] << 8));
}

void read_block(char * block, int block_number)
{
  fseek(codefile, block_number * COD_BLOCK_SIZE, SEEK_SET);
  fread(block, COD_BLOCK_SIZE, 1, codefile);
}

// reverse the endian type (big to little or little to big)
// note: there's probably a wonderful macro nestled deep in
// the bowels of the os that nicely handles this conversion...

unsigned int endian_swap(unsigned int a)
{

  return
    ( (a>>24) & 0x000000ff) | 
    ( (a>>8)  & 0x0000ff00) |
    ( (a<<8)  & 0x00ff0000) |
    ( (a<<24) & 0xff000000);

}

//-----------------------------------------------------------
// cod_address_in_range - check to see if an address falls into
// one of the valid code areas. A code area is define by a start
// address and an end address. If the address is in a valid area
// then a '1' is returned.
//

int cod_address_in_range(char *range_block,int address)
{
  int i =0;
  int start,end;

  do
    {
      // get the start and end addresses of this range

      start = get_short_int(&range_block[i])/2;
      i+=2;

      end = get_short_int(&range_block[i])/2;
      i+=2;

      if((address>=start) && (address<=end))
	return 1;  // in range

      // The end address can be zero on only the first 
      // start/end pair.

      if((end == 0) && (i>4))
	return 0;

    }
  while(i<COD_BLOCK_SIZE);

  return 0;
}

//-----------------------------------------------------------
// read_hex_from_cod - this routine will get the opcodes from
// the .cod file and intialize the pic program memory with them.
//

void read_hex_from_cod(int *memory, int memory_size)
{

  int i,j,index,address;
  char range_block[COD_BLOCK_SIZE];

  i = get_short_int(&directory_block_data[COD_DIR_MEMMAP]);
  j = get_short_int(&directory_block_data[COD_DIR_MEMMAP+2]);

  if( (i!=j) || (i==0))
    {
      cout << ".cod range error \n";
      return;
    }

  read_block(range_block, i);

  // Loop through all of the .cod file blocks that (may) contain code

  for(i=0; i<COD_CODE_IMAGE_BLOCKS; i++)
    {
      index = get_short_int(&directory_block_data[2*(COD_DIR_CODE + i)]);

      if (index != 0) 
	{
	  read_block(temp_block, index);
	  for(j=0; j<COD_BLOCK_SIZE/2; j++)
	    {
	      address = i*COD_BLOCK_SIZE/2 + j;
	      if(address < memory_size)
		{
		  if(cod_address_in_range(range_block,address))
		    memory[address] = (int)get_short_int(&temp_block[j*2]);
		  else if(address == CONFIG_ADDRESS)
		    config_word = (int)get_short_int(&temp_block[j*2]);
		}
	    }
	}
    }
}

//-----------------------------------------------------------
// Determing the .lst file name from the cod file file name.
// imo this is cheezy because the .cod file and .lst file have
// to have the same base file name. By convention, mpasm always
// made sure this happened. gpasm otoh, gives you an option to
// make the two different. Furthermore, gpasm includes the .lst 
// file file name in the list of source files within the .cod
// file - unfortunately mpasm doesn't ... so gpsim has to assume
// the list file isn't present

int cod_open_lst(char *filename)
{
  char *pc;
  int i;

  strcpy(lstfilename, filename);
  pc = strrchr(lstfilename, '.');
  if (pc == NULL) {
    if( (i = strlen(lstfilename)) < (256-4))
      pc = lstfilename + i;
    else
      return COD_FILE_NAME_TOO_LONG;
      
  }
  strcpy(pc, ".lst");

  //  if(NULL == (lstfp = fopen(lstfilename,"r")))
  //    return COD_LST_NOT_FOUND;

  return COD_SUCCESS;
}

//-----------------------------------------------------------
void read_src_files_from_cod(pic_processor *cpu)
{
#define FILE_SIZE  64
#define FILES_PER_BLOCK (COD_BLOCK_SIZE/FILE_SIZE)

  int i,j,start_block,end_block,offset,num_files;
  char b[FILE_SIZE];

  if(cpu->files) 
    {
      cout << "This cpu already has source files\n";
      return;
    }

  num_files = 0;
  start_block = get_short_int(&directory_block_data[COD_DIR_NAMTAB]);

  // First, just count the number of source files
  if(start_block) {
    end_block   = get_short_int(&directory_block_data[COD_DIR_NAMTAB+2]);
    for(j=start_block; j<=end_block; j++) {
      read_block(temp_block, j);
      for(i=0; i<FILES_PER_BLOCK; i++) {
	offset = i*FILE_SIZE;
	if(temp_block[offset])
	  num_files++;
      }
    }
  }

  cout << "found " <<num_files<< " source files in .cod file\n";

  if(num_files) {

    cpu->files = new file_context[num_files+1];
    cpu->number_of_source_files = num_files;
    cpu->lst_file_id = num_files+1;

    num_files = 0;  // now use 'num_files' as a counter.

    for(j=start_block; j<=end_block; j++) {
      read_block(temp_block, j);
      for(i=0; i<FILES_PER_BLOCK; i++) {
	offset = i*FILE_SIZE;
	substr(b,&temp_block[offset+1],FILE_SIZE);
	if(temp_block[offset]) {
	  cpu->files[num_files].name = new char [strlen(b)];
	  strcpy(cpu->files[num_files].name,b);
	  printf("%s\n",cpu->files[num_files].name);
	  cpu->files[num_files].file_ptr = fopen(cpu->files[num_files].name,"r");
	  cpu->files[num_files].max_line = 0;

	  if((strncmp(lstfilename, b,256) == 0) && 
	     (cpu->lst_file_id > cpu->number_of_source_files))
	    cpu->lst_file_id = num_files;

	  num_files++;
	}
      }
      if(num_files != cpu->number_of_source_files)
	cout << "warning, number of sources changed while reading code (gpsim bug)\n";
      if(cpu->lst_file_id > cpu->number_of_source_files)
	{
	  cpu->number_of_source_files = cpu->lst_file_id;
	  cpu->files[cpu->number_of_source_files].name = new char [strlen(lstfilename)];
	  strcpy(cpu->files[cpu->number_of_source_files].name,lstfilename);
	  printf("%s wasn't in .cod\n",cpu->files[cpu->number_of_source_files].name);
	  //cpu->files[num_files].file_ptr = fopen(cpu->files[num_files].name,"r");
	  cpu->files[num_files].max_line = 0;
	}

    }
  }else
    printf("No source file info\n");

}

//-----------------------------------------------------------
void read_line_numbers_from_cod(pic_processor *cpu)
{
  int lst_line_number = 0;
  int last_src_line = 0;
  char buf[256];
  int i,j,start_block,end_block,offset;
  int address,file_id, sline,smod;

  start_block = get_short_int(&directory_block_data[COD_DIR_LSTTAB]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_LSTTAB+2]);

    // Loop through all of the .cod file blocks that contain line number info

    for(j=start_block; j<=end_block; j++) {

      read_block(temp_block,j);

      // Get the line number info from within one .cod block

      for(offset=0; offset<COD_BLOCK_SIZE; offset += COD_LS_SIZE) {

	if((temp_block[offset+COD_LS_SMOD] & 4) == 0) {
	  lst_line_number++;
	  file_id = temp_block[offset+COD_LS_SFILE];
	  address = get_short_int(&temp_block[offset+COD_LS_SLOC]);
	  sline   = get_short_int(&temp_block[offset+COD_LS_SLINE]);
	  smod    = temp_block[offset+COD_LS_SMOD] & 0xff;

	  if( (file_id <= cpu->number_of_source_files) &&
	      (address <= cpu->program_memory_size()) &&
	      (smod == 0x80) ) {
	    cpu->attach_src_line(address,file_id,sline,lst_line_number);

	  }
	}
      }
    }
    cpu->read_src_files();
  }

}

//-----------------------------------------------------------
// open_cod_file
//
void read_symbols( pic_processor *cpu )
{
  char *s,length;
  short type;
  int i,j,start_block,end_block, value;
  char b[256];

  start_block = get_short_int(&directory_block_data[COD_DIR_LSYMTAB]);

  if(start_block) {

    end_block   = get_short_int(&directory_block_data[COD_DIR_LSYMTAB+2]);

    for(j=start_block; j<=end_block; j++) {

      read_block(temp_block, j);

      for(i=0; i<COD_BLOCK_SIZE;) {
	s =  &temp_block[i];

	if(*s==0)
	  break;

	length = *s;
	type  = *(short *)&s[length+1];
	if(type>128)
	  type = COD_ST_CONSTANT;
	value = *(int *)&s[length+3];
	value = endian_swap(value);

	switch(type) 
	  {
	  case COD_ST_C_SHORT:
	    symbol_table.add_register(substr(b,&s[1],length),
				      cpu->registers[value]);
				      
	    break;

	  case COD_ST_ADDRESS:
	    symbol_table.add_address(substr(b,&s[1],length), value);
	    break;
	    //COD_ST_CONSTANT:
	  default:
	    symbol_table.add_constant(substr(b,&s[1],length),value);
	  }

	/*	printf("%s = %x, type = %d\n",
	       substr(b,&s[1],length),
	       value,
	       type);
	*/

	i += (length + 7);
      }
    }
  }else
      printf("No long symbol table info\n");

}

//-----------------------------------------------------------
// open_cod_file
//
//  The purpose of this function is to process a .cod symbol file.
// If a cpu hasn't been declared prior to calling this function, then this
// function will attempt to determine the cpu from the .cod file.
//

int open_cod_file(pic_processor **pcpu, char *filename)
{
  int buffer_size;
  int suspicions = 0; // count the number of legal but suspicious items in the .cod file
  char processor_name[16],*pc;
  int error_code= COD_SUCCESS;

  cout << "processing cod file " << filename << '\n';

  codefile = fopen(filename,"r");

  if(codefile == NULL) {
    printf("Unable to open %s\n",filename);
    return COD_FILE_NOT_FOUND;
  }

  cod_open_lst(filename);

  /* Start off by reading the directory block */

  directory_block_data = new char[COD_BLOCK_SIZE];
  temp_block = new char[COD_BLOCK_SIZE];

  buffer_size = fread(directory_block_data, 1, COD_BLOCK_SIZE, codefile);
  if(buffer_size != COD_BLOCK_SIZE)
    {
      delete directory_block_data;
      delete temp_block;
      fclose(codefile);
      return COD_BAD_FILE;
    }

  // Perform a series of integrity checks

  // If we get here, then the .cod file is good.

  if(*pcpu == NULL) 
    {
      cout << "ascertaining cpu from the .cod file\n";
      substr(processor_name,&directory_block_data[COD_DIR_PROCESSOR],8);
      processor_name[8] = 0;
      cout << "found a " << processor_name << '\n';

      *pcpu = add_processor(processor_name,processor_name);
      if(*pcpu == NULL) return(COD_UNRECOGNIZED_PROCESSOR);

      int maxaddr =  (*pcpu)->program_memory_size();
      int memory[maxaddr];
      int i;
      for(i=0; i<maxaddr; i++)
	memory[i] = 0xffffffff;

      read_hex_from_cod(memory, maxaddr);
      (*pcpu)->build_program_memory(memory, 0, maxaddr);
      (*pcpu)->set_config_word(config_word);
    }
  else
    {
      cout << "cpu is non NULL\n";
    }

  // Read the source to list to assembly cross reference
  read_src_files_from_cod(*pcpu);

  // Associate the .lst and .asm files' line numbers with 
  // the assembly instructions' addresses.

  read_line_numbers_from_cod(*pcpu);
  read_symbols(*pcpu);

  delete directory_block_data;
  delete temp_block;
  fclose(codefile);
  return error_code;

}


void display_symbol_file_error(int err)
{

  switch(err) {

  case COD_SUCCESS:
    cout << "no error\n";
    break;

  case COD_FILE_NOT_FOUND:
    cout << "unable to find the symbol file\n";
    break;

  case COD_UNRECOGNIZED_PROCESSOR:
    cout << "unrecognized processor in the symbol file\n";
    break;

  case COD_BAD_FILE:
    cout << "bad file format\n";
    break;
  }
}
