/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#ifndef __PIC_INSTRUCTIONS_H__
#define __PIC_INSTRUCTIONS_H__


/*
 *  base class for an instruction
 */

class instruction
{
public:

  enum INSTRUCTION_TYPES
  {
    NORMAL_INSTRUCTION,
    INVALID_INSTRUCTION,
    BREAKPOINT_INSTRUCTION
  };

  /*
   * Not all instructions derived from the instruction
   * class use these constants... 
   */

  enum
  {
    REG_MASK_12BIT = 0x1f,
    REG_MASK_14BIT = 0x7f,
    DESTINATION_MASK_12BIT = 0x20,
    DESTINATION_MASK_14BIT = 0x80
  };


  char name_str[20];      /* %%% FIX ME %%% dyanmically allocate? */
  unsigned int opcode;
  pic_processor *cpu;     /* A pointer to the microcontroller to which this
                           *  instruction belongs  */
  int file_id;            /* The source file that declared this instruction
			   * (The file_id is an index into an array of files) */
  int src_line;           /* The line number within the source file */
  int lst_line;           /* The line number within the list file */

  virtual void execute(void){ };
  virtual void debug(void){ };
  virtual char *name(char *str){ return(name_str);};
  virtual int get_src_line(void) { return(src_line); };
  virtual int get_lst_line(void) { return(lst_line); };
  virtual INSTRUCTION_TYPES isa(void) {return NORMAL_INSTRUCTION;};
  void decode(pic_processor *new_cpu, unsigned int new_opcode){cpu = new_cpu;opcode=new_opcode;};
  void add_line_number_symbol(int address);
  void update_line_number(int file, int sline, int lline);
};


//---------------------------------------------------------
class invalid_instruction : public instruction
{

  virtual void execute(void)
  { 
    cout << "*** INVALID INSTRUCTION ***\n";
    #ifdef __DEBUG_VERBOSE__
      debug();
    #endif
  };
  virtual void debug(void)
  {
      cout << "*** INVALID INSTRUCTION ***\n";
  };

  virtual INSTRUCTION_TYPES isa(void) {return INVALID_INSTRUCTION;};
  virtual char *name(char *str){return("INVALID");};
};

//---------------------------------------------------------
class Literal_op : public instruction
{
public:
  unsigned int L;

  virtual void execute(void){ };
  virtual void debug(void){ };
  virtual char *name(char *);

  void decode(pic_processor *new_cpu, unsigned int new_opcode)
    {
      opcode = new_opcode;
      cpu = new_cpu;
      L = opcode & 0xff;
    }
};


//---------------------------------------------------------
class Bit_op : public instruction
{
public:
  unsigned int mask,register_address;
  file_register *reg;

  virtual void execute(void){ };
  virtual void debug(void){ };
  virtual char *name(char *);

  void decode(pic_processor *new_cpu, unsigned int new_opcode);

};


//---------------------------------------------------------
class Register_op : public instruction
{
public:

  static file_register *source;
  unsigned int register_address;
  bool destination;

  /*  file_register *destination;*/

  virtual void execute(void){ };
  virtual void debug(void){ };
  virtual char *name(char *);

  void decode(pic_processor *new_cpu, unsigned int new_opcode);

};




//---------------------------------------------------------
extern invalid_instruction bad_instruction;



//---------------------------------------------------------
class ADDWF : public Register_op
{
public:

  ADDWF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------

class ANDLW : public Literal_op
{

public:
  ANDLW(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class ANDWF : public Register_op
{
public:

  ANDWF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class BCF : public Bit_op
{
public:

  BCF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class BSF : public Bit_op
{
public:

  BSF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class BTFSC : public Bit_op
{
public:

  BTFSC(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class BTFSS : public Bit_op
{
public:

  BTFSS(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class CALL : public instruction
{
public:
  unsigned int destination;

  CALL(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);
  char *name(char *str);
};

//---------------------------------------------------------
class CLRF : public Register_op
{
public:

  CLRF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);
  char *name(char *str);
};

//---------------------------------------------------------
class CLRW : public instruction
{
public:

  CLRW(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class CLRWDT : public instruction
{
public:

  CLRWDT(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class COMF : public Register_op
{
public:

  COMF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class DECF : public Register_op
{
public:

  DECF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class DECFSZ : public Register_op
{
public:

  DECFSZ(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class GOTO : public instruction
{
public:
  unsigned int destination;

  GOTO(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);
  char *name(char *str);
};

//---------------------------------------------------------
class INCF : public Register_op
{
public:

  INCF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class INCFSZ : public Register_op
{
public:

  INCFSZ(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------

class IORLW : public Literal_op
{

public:
  IORLW(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class IORWF : public Register_op
{
public:

  IORWF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class MOVF : public Register_op
{
public:

  MOVF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);
  virtual void debug(void);

};

//---------------------------------------------------------

class MOVLW : public Literal_op
{
public:
  MOVLW(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class MOVWF : public Register_op
{
public:

  MOVWF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);
  char *name(char *str);
};

//---------------------------------------------------------
class NOP : public instruction
{
public:

  NOP(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class OPTION : public instruction
{
public:

  OPTION(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------

class RETLW : public Literal_op
{
public:

  RETLW(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class RLF : public Register_op
{
public:

  RLF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class RRF : public Register_op
{
public:

  RRF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};


//---------------------------------------------------------
class SLEEP : public instruction
{
public:

  SLEEP(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class SUBWF : public Register_op
{
public:

  SUBWF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};


//---------------------------------------------------------
class SWAPF : public Register_op
{
public:

  SWAPF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};


//---------------------------------------------------------
class TRIS : public Register_op
{
public:
  file_register *reg;

  TRIS(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);
  char *name(char *str);
};

//---------------------------------------------------------

class XORLW : public Literal_op
{

public:

  XORLW(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};

//---------------------------------------------------------
class XORWF : public Register_op
{
public:

  XORWF(pic_processor *new_cpu, unsigned int new_opcode);
  virtual void execute(void);

};


#endif  /*  __PIC_INSTRUCTIONS_H__ */
